#include "matrixproduct.hpp"
#include "bprop.hpp"
#include "wchar.h"
#include "string"

/**
 Compute the messages using the specified method. Essentially just a wrapper function around each of the algorithms.

 A, B: input matrices.
 C: output matrix.
 dir: message passing direction.
 N: matrix size.
 nread: returns the number of addition operations.

 The code in algos.cpp finds min_b A(a,b) + B(b,c)
 */
void passmessage_helper(Scalar** A, Scalar** B, Scalar** C, int dir, int method, int N, Scalar** res, long long& nread)
{
  Scalar** trans = new Scalar*[N];
  for (int i = 0; i < N; i++)
  {
    /*
     Transpose the matrices appropriately to be compatible with the code in algos.cpp

     m(a,b) = A(a,b) + min_c { B(a,c) + C(b,c) }
     -> A(a,b) + min_c { B(a,c) + C'(c,b) }
     m(b,c) = C(b,c) + min_a { A(a,b) + B(a,c) }
     -> C(b,c) + min_a { A'(b,a) + B(a,c) }
     */
    trans[i] = new Scalar[N];
    if (dir == LEFT)
      for (int j = 0; j < N; j++)
        trans[i][j] = C[j][i];
    else
      for (int j = 0; j < N; j++)
        trans[i][j] = A[j][i];
  }

  int eval;
  Scalar** minmarg = NULL;

  Scalar** M0;
  Scalar** M1;
  Scalar** M2;

  if (dir == LEFT)
  {
    M0 = A;
    M1 = B;
    M2 = trans;
  }
  else
  {
    M0 = C;
    M1 = trans;
    M2 = B;
  }

  // Pass the matrices to the appropriate algorithm.
  if (method == NAIVE)
  {
    minmarg = slow(M1, M2, N);
    eval = N * N * N;
  }
  else
  {
    minmarg = sorting(M1, M2, method, N, eval);
  }

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
      res[i][j] = M0[i][j] PLUS minmarg[i][j];

  for (int i = 0; i < N; i++)
  {
    delete[] trans[i];
    delete[] minmarg[i];
  }
  delete[] trans;
  delete[] minmarg;

  nread += eval;
}

int model::addedge(potential* p, int n1, int n2)
{
  edges.push_back(pair<int, int> (n1, n2));
  pots.push_back(p);
  return pots.size() - 1;
}

void model::propagate(vector<pair<int, int> >* messageorder, long long& nread)
{
  for (vector<pair<int, int> >::iterator it = messageorder->begin(); it != messageorder->end(); it++)
  {
    int from = it->first;
    int to = it->second;

    int direction_from = LEFT;
    int direction_to = LEFT;

    if (edges[from].first == edges[to].first)
    {
      direction_from = LEFT;
      direction_to = LEFT;
    }
    if (edges[from].first == edges[to].second)
    {
      direction_from = LEFT;
      direction_to = RIGHT;
    }
    if (edges[from].second == edges[to].first)
    {
      direction_from = RIGHT;
      direction_to = LEFT;
    }
    if (edges[from].second == edges[to].second)
    {
      direction_from = RIGHT;
      direction_to = RIGHT;
    }
    Scalar* mess = pots[from]->send(to, direction_from, nread);
    pots[to]->receive(from, direction_to, mess);
  }
}

int model::MAP(int edgeid, int direction, long long& nread)
{
  return pots[edgeid]->MAP(direction, nread);
}

