#ifndef BPROP_HPP
#define BPROP_HPP

/// All belief-propagation routines for higher-order models.

#include "common.hpp"
#include "bprop.hpp"
#include "maximize.hpp"
#include "potential.hpp"

void passmessage_helper(Scalar** A, Scalar** B, Scalar** C, int dir, int method, int N, Scalar** res, long long& nread);

class model
{
  public:
    vector<pair<int, int> > edges;
    vector<potential*> pots;

    int addedge(potential* p, int n1, int n2);

    void propagate(vector<pair<int, int> >* messageorder, long long& nread);
    int MAP(int edgeid, int direction, long long& nread);
};

/** A clique, and its messages. */
class cliquemessage
{
  public:
    Scalar** A;
    Scalar** B;
    Scalar** C;
    Scalar** mleft;
    Scalar** mright;
    int N;

    cliquemessage(Scalar** A, Scalar** B, Scalar** C, int N) :
      A(A), B(B), C(C), N(N)
    {
      mleft = new Scalar*[N];
      mright = new Scalar*[N];
      for (int i = 0; i < N; i++)
      {
        mleft[i] = new Scalar[N];
        mright[i] = new Scalar[N];
        for (int j = 0; j < N; j++)
        {
          mleft[i][j] = SEMIRING_ZERO;
          mright[i][j] = SEMIRING_ZERO;
        }
      }
    }
    ;
    ~cliquemessage()
    {
      for (int i = 0; i < N; i++)
      {
        delete[] mleft[i];
        delete[] mright[i];
        delete[] A[i];
        delete[] B[i];
        delete[] C[i];
      }
      delete[] mleft;
      delete[] mright;
      delete[] A;
      delete[] B;
      delete[] C;
    }
    ;

    /** Pass a message LEFT or RIGHT. */
    void passmessage(int dir, int method, Scalar** res, long long& nread)
    {
      Scalar** Ap = new Scalar*[N];
      Scalar** Cp = new Scalar*[N];
      for (int i = 0; i < N; i++)
      {
        Ap[i] = new Scalar[N];
        Cp[i] = new Scalar[N];

        for (int j = 0; j < N; j++)
        {
          Ap[i][j] = A[i][j];
          if (dir == RIGHT) Ap[i][j] = Ap[i][j] PLUS mleft[i][j];
          Cp[i][j] = C[i][j];
          if (dir == LEFT) Cp[i][j] = Ap[i][j] PLUS mright[i][j];
        }
      }
      passmessage_helper(Ap, B, Cp, dir, method, N, res, nread);

      for (int i = 0; i < N; i++)
      {
        delete[] Ap[i];
        delete[] Cp[i];
      }
      delete[] Ap;
      delete[] Cp;
    }

    /** Compute the maximum likelihood state, given our current messages. */
    void MAP(int dir, int method, Scalar** space, int* res, long long& nread)
    {
      Scalar** Ap = new Scalar*[N];
      Scalar** Cp = new Scalar*[N];
      for (int i = 0; i < N; i++)
      {
        Ap[i] = new Scalar[N];
        Cp[i] = new Scalar[N];

        for (int j = 0; j < N; j++)
        {
          Ap[i][j] = A[i][j] PLUS mleft[i][j];
          Cp[i][j] = C[i][j] PLUS mright[i][j];
        }
      }
      passmessage_helper(Ap, B, Cp, dir, method, N, space, nread);
      res[0] = 0;
      res[1] = 0;
      for (int i = 0; i < N; i++)
        for (int j = 0; j < N; j++)
          if (space[i][j] COMP space[res[0]][res[1]])
          {
            res[0] = i;
            res[1] = j;
          }
      for (int i = 0; i < N; i++)
      {
        delete[] Ap[i];
        delete[] Cp[i];
      }
      delete[] Ap;
      delete[] Cp;
    }
};

/** A graphical model. */
class gmodel
{
  public:
    vector<cliquemessage*> vc;
    int method;
    gmodel(vector<cliquemessage*> vc, int method) :
      vc(vc), method(method)
    {
    }
    ;
    ~gmodel()
    {
    }
    ;

    /** Pass messages in a chain (for the text-denoising experiments). */
    void passmessage_chain(long long& nread)
    {
      for (int i = 0; i < (int) vc.size() - 1; i++)
        vc[i]->passmessage(RIGHT, method, vc[i + 1]->mleft, nread);
      for (int i = vc.size() - 1; i > 0; i--)
        vc[i]->passmessage(LEFT, method, vc[i - 1]->mright, nread);
    }

    /** Pass messages in a ring (for the random and graph-matching experiment). */
    void passmessage_ring(long long& nread)
    {
      vector<pair<int, int> > ms;
      int N = vc.size();
      for (int i = 0; i < N; i++)
      {
        ms.push_back(pair<int, int> (i, LEFT));
        ms.push_back(pair<int, int> (i, RIGHT));
      }
      random_shuffle(ms.begin(), ms.end());
      for (vector<pair<int, int> >::iterator it = ms.begin(); it != ms.end(); it++)
      {
        int n1 = it->first;
        int dir = it->second;
        if (dir == LEFT)
          vc[n1]->passmessage(LEFT, method, vc[(N + n1 - 1) % N]->mright, nread);
        else
          vc[n1]->passmessage(RIGHT, method, vc[(n1 + 1) % N]->mleft, nread);
      }
    }

    /** Compute the maximum likelihood state of the entire model. */
    void MAP(int dir, int N, int* res, int I, long long& nread)
    {
      Scalar** space = new Scalar*[N];
      for (int i = 0; i < N; i++)
        space[i] = new Scalar[N];
      vc[I]->MAP(dir, method, space, res, nread);
      for (int i = 0; i < N; i++)
        delete[] space[i];
      delete[] space;
    }
};

#endif
