#include "experiment.hpp"
#include "matrixproduct.hpp"
#include "bprop.hpp"
#include "potential.hpp"
#include "grid.hpp"

#include "sstream"
#include "sys/stat.h"

int* gmatch_inner(int method, int ran, vector<point<Scalar> > template_graph, vector<point<Scalar> > target_graph, int iterations, double& time, long long& nread, Scalar& energy)
{
  energy = 0;
  long T = clock();
  int M = template_graph.size();
  int N = target_graph.size();

  vector<cliquemessage*> cliques;
  // Set up the potentials in a ring, connecting 2-nearest-neighbours.
  for (int m = 0; m < M; m++)
  {

    int p1 = m;
    int p2 = (m + 1) % M;
    int p3 = (m + 2) % M;
    Scalar d1 = template_graph[p1].distance1(template_graph[p2]);
    Scalar d2 = template_graph[p1].distance1(template_graph[p3]);
    Scalar** A = new Scalar*[N];
    Scalar** B = new Scalar*[N];
    Scalar** C = new Scalar*[N];
    for (int i = 0; i < N; i++)
    {
      A[i] = new Scalar[N];
      B[i] = new Scalar[N];
      C[i] = new Scalar[N];
      for (int j = 0; j < N; j++)
      {
        Scalar d = target_graph[i].distance1(target_graph[j]);

        if (!ran)
        {
          A[i][j] = fabs(d1 - d);
          B[i][j] = fabs(d2 - d);
          C[i][j] = 0;
        }
        else
        {
          A[i][j] = rand() / (1.0 * RAND_MAX);
          B[i][j] = rand() / (1.0 * RAND_MAX);
          C[i][j] = rand() / (1.0 * RAND_MAX);
        }
      }
    }
    cliques.push_back(new cliquemessage(A, B, C, N));
  }
  gmodel G(cliques, method);

  // Pass messages using loopy-BP.
  for (int i = 0; i < iterations; i++)
    G.passmessage_ring(nread);

  int* res = new int[M];
  for (int i = 0; i < M - 1; i++)
    G.MAP(LEFT, N, res + i, i, nread);

  energy = SEMIRING_ZERO;
  for (int i = 0; i < M; i++)
  {
    cliquemessage* cm = cliques[i];
    energy = energy PLUS cm->A[res[i]][res[(i + 1) % M]] PLUS cm->B[res[i]][res[(i + 2) % M]] PLUS cm->C[res[(i + 1) % M]][res[(i + 2) % M]];
  }

  for (int i = 0; i < (int) cliques.size(); i++)
    delete cliques[i];
  time = ((double) (clock() - T)) / CLOCKS_PER_SEC;
  return res;
}

/** Graph matching / random experiment...
 * The random version of the experiment uses the same graph topology, but initialises the potentials randomly.
 */
int* gmatch(int method, int M, int N, double& time, int random, int iterations, long long& nread, Scalar& energy)
{
  nread = 0;
  vector<point<Scalar> > template_graph;
  vector<point<Scalar> > target_graph;
  //int M = 10;
  assert(N >= M);

  for (int i = 0; i < N; i++)
  {
    point<Scalar> p((rand() % 10000) / 10000.0, (rand() % 10000) / 10000.0);
    if (i < M) template_graph.push_back(p);
    target_graph.push_back(p);
  }

  int* res = gmatch_inner(method, random, template_graph, target_graph, iterations, time, nread, energy);
  return res;
}

int* random(int method, int M, int N, double& time, long long& nread, Scalar& energy)
{
  energy = 0;
  long T = clock();
  nread = 0;

  vector<cliquemessage*> cliques;
  // Set up the potentials in a chain, connecting 2-nearest-neighbours.
  for (int m = 0; m < M - 2; m++)
  {
    Scalar** A = new Scalar*[N];
    Scalar** B = new Scalar*[N];
    Scalar** C = new Scalar*[N];
    for (int i = 0; i < N; i++)
    {
      A[i] = new Scalar[N];
      B[i] = new Scalar[N];
      C[i] = new Scalar[N];
      for (int j = 0; j < N; j++)
      {
        A[i][j] = rand() / (1.0 * RAND_MAX);
        B[i][j] = rand() / (1.0 * RAND_MAX);
        C[i][j] = rand() / (1.0 * RAND_MAX);
      }
    }
    cliques.push_back(new cliquemessage(A, B, C, N));
  }
  gmodel G(cliques, method);

  // Don't actually pass the message: if all of the messages are zero, computing the MAP states just performs matrix product.
  //G.passmessage_chain(nread);

  int* res = new int[M];
  for (int i = 0; i < M - 3; i++)
    G.MAP(LEFT, N, res + i, i, nread);
  G.MAP(RIGHT, N, res + M - 2, M - 3, nread);

  energy = SEMIRING_ZERO;
  for (int i = 0; i < M - 2; i++)
  {
    cliquemessage* cm = cliques[i];
    energy = energy PLUS cm->A[res[i]][res[(i + 1) % M]] PLUS cm->B[res[i]][res[(i + 2) % M]] PLUS cm->C[res[(i + 1) % M]][res[(i + 2) % M]];
  }

  for (int i = 0; i < (int) cliques.size(); i++)
    delete cliques[i];
  time = ((double) (clock() - T)) / CLOCKS_PER_SEC;
  return res;
}
