#include "grid.hpp"
#include "bprop.hpp"
#include "potential.hpp"
#include "util.hpp"

#include "CImg.h"
using namespace cimg_library;

int** grid(int method, Scalar*** nodepot, Scalar** smoothness, int X, int Y, int N, double& online, double& total, Scalar& energy, long long& nread)
{
  nread = 0;
  int** Pa = new int*[N];
  int** Painv = new int*[N];
  for (int i = 0; i < N; i++)
  {
    Pa[i] = new int[N];
    Painv[i] = new int[N];
  }

  // Indices of horizontal and vertical edges of the grid.
  map<pair<int, int> , int> edgeind_horizontal;
  map<pair<int, int> , int> edgeind_vertical;
  model M;

  for (int y = 0; y < Y; y++)
    for (int x = 0; x < X; x++)
    {
      // Node potentials are merged into the horizontal edges.
      if (x < X - 1)
      {
        potential* pot;
        if (x < X - 2)
          pot = new potential(smoothness, smoothness, nodepot[y][x], NULL, Pa, Pa, Painv, Painv, N, method);
        else
          pot = new potential(smoothness, smoothness, nodepot[y][x], nodepot[y][x + 1], Pa, Pa, Painv, Painv, N, method);
        edgeind_horizontal[pair<int, int> (x, y)] = M.addedge(pot, y * X + x, y * X + x + 1);
      }

      // Vertical edges only encode smoothness.
      if (y < Y - 1)
      {
        potential* pot;
        if (y < Y - 2)
          pot = new potential(smoothness, smoothness, NULL, NULL, Pa, Pa, Painv, Painv, N, method);
        else
          pot = new potential(smoothness, smoothness, NULL, NULL, Pa, Pa, Painv, Painv, N, method);
        edgeind_vertical[pair<int, int> (x, y)] = M.addedge(pot, y * X + x, (y + 1) * X + x);
      }
    }

  // Order of edges for message passing. Each entry (A,B) means a message is passed from the edge A to the edge B.
  vector<pair<int, int> > morder;

  if (Y > 1)
  { // A grid-structured model.
    for (int y = 0; y < Y; y++)
      for (int x = 0; x < X; x++)
      {
        if (x < X - 1)
        {
          int ind = edgeind_horizontal[pair<int, int> (x, y)];
          if (x < X - 2) morder.push_back(pair<int, int> (edgeind_horizontal[pair<int, int> (x + 1, y)], ind));
          if (y < Y - 1)
          {
            morder.push_back(pair<int, int> (edgeind_vertical[pair<int, int> (x, y)], ind));
            morder.push_back(pair<int, int> (edgeind_vertical[pair<int, int> (x + 1, y)], ind));
          }
          if (y > 0)
          {
            morder.push_back(pair<int, int> (edgeind_vertical[pair<int, int> (x, y - 1)], ind));
            morder.push_back(pair<int, int> (edgeind_vertical[pair<int, int> (x + 1, y - 1)], ind));
          }
        }
        if (y < Y - 1)
        {
          int ind = edgeind_vertical[pair<int, int> (x, y)];
          if (y < Y - 2) morder.push_back(pair<int, int> (edgeind_vertical[pair<int, int> (x, y + 1)], ind));
        }
      }

    int ms = morder.size();
    for (int i = 0; i < ms; i++)
    {
      pair<int, int> edge = morder[i];
      morder.push_back(pair<int, int> (edge.second, edge.first));
    }
  }
  else
  { // A chain structured model.
    for (int x = 0; x < X - 1; x++)
    {
      int ind1 = edgeind_horizontal[pair<int, int> (x, 0)];
      int ind2 = edgeind_horizontal[pair<int, int> (x + 1, 0)];
      morder.push_back(pair<int, int> (ind1, ind2));
    }
    for (int x = X - 2; x >= 1; x--)
    {
      int ind1 = edgeind_horizontal[pair<int, int> (x, 0)];
      int ind2 = edgeind_horizontal[pair<int, int> (x - 1, 0)];
      morder.push_back(pair<int, int> (ind1, ind2));
    }
  }

  clock_t time_total = clock();

  // Precompute the order statistics of the smoothness terms.
  if (method != NAIVE) for (int i = 0; i < N; i++)
    permfunction(smoothness[i], Pa[i], Painv[i], N);

  clock_t time_online = clock();
  for (int i = 0; i < (Y > 1 ? arg_iterations : 1); i++)
  {
    // Propagate messages in a random order for a grid, or in a fixed order (as defined above) for a chain.
    if (Y > 1) random_shuffle(morder.begin(), morder.end());
    M.propagate(&morder, nread);
  }
  int** res = new int*[Y];
  for (int y = 0; y < Y; y++)
    res[y] = new int[X];

  // Compute the MAP states after messages have been passed.
  for (int y = 0; y < Y; y++)
    for (int x = 0; x < X - 1; x++)
    {
      res[y][x] = M.MAP(edgeind_horizontal[pair<int, int> (x, y)], LEFT, nread);
      if (x == X - 2) res[y][X - 1] = M.MAP(edgeind_horizontal[pair<int, int> (x, y)], RIGHT, nread);
    }
  clock_t finish = clock();

  online = ((double) (finish - time_online)) / CLOCKS_PER_SEC;
  total = ((double) (finish - time_total)) / CLOCKS_PER_SEC;

  // Compute the energy of the computed solution.
  energy = nodepot[0][0][res[0][0]];
  for (int i = 0; i < Y; i++)
    for (int j = 0; j < X; j++)
      if (i != 0 or j != 0) energy = energy PLUS nodepot[i][j][res[i][j]];
  for (int i = 0; i < Y; i++)
    for (int j = 0; j < X - 1; j++)
      energy = energy PLUS smoothness[res[i][j]][res[i][j + 1]];
  for (int i = 0; i < Y - 1; i++)
    for (int j = 0; j < X; j++)
      energy = energy PLUS smoothness[res[i][j]][res[i + 1][j]];

  for (int i = 0; i < N; i++)
  {
    delete[] Pa[i];
    delete[] Painv[i];
  }
  delete[] Pa;
  delete[] Painv;

  for (int i = 0; i < (int) M.pots.size(); i++)
    delete M.pots[i];

  return res;
}
