#include "improcess.hpp"
#include "util.hpp"
#include "grid.hpp"

#include "CImg.h"
using namespace cimg_library;

int** improcess(int method, int N, int experiment, double& online, double& total, const char* imleft, const char* imright, Scalar& energy, int& X, int& Y, long long& nread)
{
  // Resize the images to 50x50.
  CImg<int> image_left = readimage(imleft, X, Y, 50);
  CImg<int> image_right = readimage(imright, X, Y, 50);

  srand(0);
  // Note: image denoising and stereo disparity are failure cases.
  if (experiment == ICML_DENOISE)
  { // Add noise...
    for (int y = 0; y < Y; y++)
      for (int x = 0; x < X; x++)
      {
        image_left(x, y) += (int) randnorm(0, 25);
        if (image_left(x, y) < 0) image_left(x, y) = 0;
        if (image_left(x, y) >= 255) image_left(x, y) = 255;
      }
  }

  Scalar** smoothness;

  int a, b;
  if (experiment == ICML_DENOISE)
  { // Extract a simple pairwise smoothness prior from a single image.
    CImg<int> t = readimage("tamarin.jpg", a, b, 0);
    smoothness = extract_pairwise(N, t);
  }
  else
    smoothness = smoothness_mat(N, experiment);

  // Construct a grid model for the image.
  Scalar*** nodepot = new Scalar**[Y];
  for (int y = 0; y < Y; y++)
  {
    nodepot[y] = new Scalar*[X];
    for (int x = 0; x < X; x++)
      nodepot[y][x] = unary_mat(x, y, image_left, image_right, N, experiment);
  }

  int** res = grid(method, nodepot, smoothness, X, Y, N, online, total, energy, nread);

  for (int y = 0; y < Y; y++)
    for (int x = 0; x < X; x++)
      if (res[y][x] == N - 1)
        res[y][x] = 255;
      else
        res[y][x] = res[y][x] * 255.0 / N;
  writeimage("output.jpg", res, X, Y, N);

  // Cleanup.
  for (int i = 0; i < N; i++)
    delete[] smoothness[i];
  delete[] smoothness;

  for (int y = 0; y < Y; y++)
  {
    for (int x = 0; x < X; x++)
      delete[] nodepot[y][x];
    delete[] nodepot[y];
  }
  delete[] nodepot;
  return res;
}

CImg<int> readimage(const char* name, int& X, int& Y, int imdim)
{
  CImg<int> image(name);

  if (imdim != 0) image = image.resize(imdim, imdim);

  X = image.width();
  Y = image.height();

  return image;
}

void writeimage(const char* name, int** data, int X, int Y, int N)
{
  CImg<int> image(X, Y, 1, 3, 0);
  for (int y = 0; y < Y; y++)
    for (int x = 0; x < X; x++)
    {
      image(x, y, 0) = data[y][x];
      image(x, y, 1) = data[y][x];
      image(x, y, 2) = data[y][x];
      if (data[y][x] == N - 1)
      {
        image(x, y, 0) = 255;
        image(x, y, 1) = 0;
        image(x, y, 2) = 0;
      }
    }

  image.save(name);
}

// Smoothness term for image processing experiments.
Scalar smoothness(int value1, int value2, int N, int experiment)
{
  int xflow1, yflow1, xflow2, yflow2;
  int rootN = (int) sqrt(N);
  switch (experiment)
  {
    case ICML_RANDOM:
      return rand() % N;
    case ICML_DISPARITY:
      return 1000 * fabs(value1 - value2) / N;
    case ICML_DENOISE: // unused.
      return 10 * fabs(value1 - value2) * fabs(value1 - value2) / N;
    case ICML_FLOW:
      xflow1 = value1 % rootN - rootN / 2;
      yflow1 = value1 / rootN - rootN / 2;
      xflow2 = value2 % rootN - rootN / 2;
      yflow2 = value2 % rootN - rootN / 2;
      return fabs(xflow1 - xflow2) + fabs(yflow1 - yflow2);
    default:
      return -1;
  }
}

// Unary term for image processing experiments.
Scalar unary(int x, int y, int value, CImg<int>& left, CImg<int>& right, int N, int experiment)
{
  double disp, intensity, pixel1, pixel2;
  int xflow, yflow;
  int rootN = (int) sqrt(N);
  switch (experiment)
  {
    case ICML_RANDOM:
      return rand() % N;
    case ICML_DISPARITY:
      disp = value * 10.0 / N;
      if (value == N - 1) return 10000;
      if (x - disp < 0) return 10000000;
      pixel1 = left(x, y) / 255.0;
      pixel2 = right(x - disp, y) / 255.0;
      return 1000 * fabs(pixel1 - pixel2);
    case ICML_DENOISE:
      intensity = value * 1.0 / N;
      return 100 * fabs(left(x, y) / 255.0 - intensity);
    case ICML_FLOW:
      xflow = value % rootN - rootN / 2;
      yflow = value / rootN - rootN / 2;
      if (x + xflow < 0 or x + xflow >= (int) right.width() or y + yflow < 0 or y + yflow >= (int) right.height()) return 10000000;
      pixel1 = left(x, y);
      pixel2 = right(x + xflow, y + yflow);
      return 1000 * fabs(pixel1 - pixel2);
    default:
      return -1;
  }
}

// Extract the pairwise statistics of an image.
Scalar** extract_pairwise(int N, CImg<int>& im)
{
  Scalar** pairwise = new Scalar*[N];
  for (int i = 0; i < N; i++)
  {
    pairwise[i] = new Scalar[N];
    for (int j = 0; j < N; j++)
      pairwise[i][j] = 0;
  }

  int X = im.width();
  int Y = im.height();

  for (int y = 0; y < Y; y++)
    for (int x = 0; x < X; x++)
    {
      if (y < Y - 1)
      {
        int v1 = (int) (im(x, y) / 256.0 * N);
        int v2 = (int) (im(x, y + 1) / 256.0 * N);
        assert(v1 < N and v2 < N);
        pairwise[v1][v2]--;
        pairwise[v2][v1]--;
      }
      if (x < X - 1)
      {
        int v1 = (int) (im(x, y) / 256.0 * N);
        int v2 = (int) (im(x + 1, y) / 256.0 * N);
        assert(v1 < N and v2 < N);
        pairwise[v1][v2]--;
        pairwise[v2][v1]--;
      }
    }

  return pairwise;
}

Scalar** smoothness_mat(int N, int experiment)
{
  Scalar** res = new Scalar*[N];
  for (int i = 0; i < N; i++)
  {
    res[i] = new Scalar[N];
    for (int j = 0; j < N; j++)
      res[i][j] = smoothness(i, j, N, experiment);
  }
  return res;
}

Scalar* unary_mat(int x, int y, CImg<int>& left, CImg<int>& right, int N, int experiment)
{
  Scalar* res = new Scalar[N];
  for (int i = 0; i < N; i++)
    res[i] = unary(x, y, i, left, right, N, experiment);
  return res;
}

