#include "common.hpp"
#include "experiment.hpp"
#include "improcess.hpp"
#include "matrixproduct.hpp"
#include "textprocess.hpp"
#include "grid.hpp"
#include "util.hpp"
#include "CImg.h"
using namespace cimg_library;

int arg_experiment = GMODEL_GMATCH; // The graph-matching experiment
int arg_iterations = 10; // The number of iterations if loopy BP is being used
int arg_method = FAST; // The method from the AISTATS/ICML papers
int arg_nodes = 20; // The model size
int arg_vector_length = 100; // The number of states per node, where it is a variable

void deleteGrid(int** res, int Y)
{
  for (int y = 0; y < Y; y++)
    delete[] res[y];
  delete[] res;
}

void run_experiment(int experiment, int method)
{
  double time, energy;
  long long nread = 0;
  int* res = NULL;

  if (experiment == GMODEL_GMATCH || experiment == GMODEL_GMATCH_RANDOM || experiment == GMODEL_RANDOM)
  {
    srand(0);
    if (experiment == GMODEL_GMATCH || experiment == GMODEL_GMATCH_RANDOM)
    {
      int random = (experiment == GMODEL_GMATCH_RANDOM ? 1 : 0);
      res = gmatch(method, arg_nodes, arg_vector_length, time, random, arg_iterations, nread, energy);
    }
    else
    {
      res = random(method, arg_nodes, arg_vector_length, time, nread, energy);
    }
    printf("time=%.2f; multiplications=%lld\n", time, nread);
    delete[] res;
  }
  else if (experiment == ICML_TEXT || experiment == GMODEL_TEXT)
  {
    std::string names[15] =
    { "cat", "de05_", "dk", "ee", "en", "fi", "fr05_", "it", "nl", "no", "se", "sorb", "tr", "kr", "jp" };
    std::string encs[15] =
    { "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-4", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-1", "r,ccs=ISO-8859-2", "r,ccs=ISO-8859-9", "r,ccs=UTF-8", "r,ccs=UTF-8" };

    int trainlen = 100000;
    int skipchain = (experiment == GMODEL_TEXT ? 1 : 0);
    int testlen = arg_nodes;
    int nCorpora = (skipchain ? 14 : 15); // Japanese is too memory intensive for skipchain

    // Run text experiment for each language corpus
    for (int i = 0; i < nCorpora; i++)
    {
      char* corpus = new char[256];
      sprintf(corpus, "../../corpora/%s100k/sentences.txt", names[i].c_str());
      int perf, errs;

      float typoProb = 0.05;
      int N = alphabet_size(corpus, trainlen, testlen, encs[i].c_str());

      srand(0);
      res = textprocess(method, corpus, trainlen, testlen, encs[i].c_str(), time, nread, typoProb, skipchain, errs, perf, energy);

      printf("alphabet size=%d; time=%.2f; multiplications=%lld\n", N, time, nread);

      delete[] corpus;
      delete[] res;
    }
  }
  else if (experiment == ICML_FLOW || experiment == ICML_DISPARITY || experiment == ICML_DENOISE || experiment == ICML_RANDOM)
  {

    std::string imleft = "stereo-left.jpg";
    std::string imright = "stereo-right.jpg";
    if (experiment == ICML_FLOW)
    {
      std::string imleft = "sphere.1.jpg";
      std::string imright = "sphere.2.jpg";
    }
    double online, total, energy;

    srand(0);
    int X, Y;
    int** res = improcess(method, arg_vector_length, experiment, online, total, imleft.c_str(), imright.c_str(), energy, X, Y, nread);

    printf("online time=%f; time=%f; multiplications=%lld\n", online, total, nread);

    deleteGrid(res, Y);
  }
}

int main(int argc, char** argv)
{
   srand(0);
   printf("# testing...\n");
   for (int i = 0; i < 100; i ++) test(100, FAST);
   printf("Tested matrix product computation.\n");

  // Parse commandline arguments
  for (int i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--experiment") && i + 1 < argc) // Which experiment should we run?
    {
      i++;
      if (!strcmp(argv[i], "gmodel-gmatch")) // Graph matching
        arg_experiment = GMODEL_GMATCH;
      else if (!strcmp(argv[i], "gmodel-gmatch-random")) // Random potentials on the above model
        arg_experiment = GMODEL_GMATCH_RANDOM;
      else if (!strcmp(argv[i], "gmodel-random")) // Random matrix products
        arg_experiment = GMODEL_RANDOM;
      else if (!strcmp(argv[i], "gmodel-text")) // Skip-chain text correction
        arg_experiment = GMODEL_TEXT;
      else if (!strcmp(argv[i], "icml-denoise")) // Image denoising
        arg_experiment = ICML_DENOISE;
      else if (!strcmp(argv[i], "icml-disparity")) // Stereo disparity
        arg_experiment = ICML_DISPARITY;
      else if (!strcmp(argv[i], "icml-flow")) // Optical flow
        arg_experiment = ICML_FLOW;
      else if (!strcmp(argv[i], "icml-random")) // Random potentials on a grid
        arg_experiment = ICML_RANDOM;
      else if (!strcmp(argv[i], "icml-text")) // Text correction on a chain
        arg_experiment = ICML_TEXT;
      else
      {
        printf("# ERROR: invalid experiment '%s'\n", argv[i]);
        exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "--iterations") && i + 1 < argc) // How many iterations should we use (if it's a loopy experiment)?
    {
      arg_iterations = atoi(argv[++i]);
      assert(arg_iterations > 0);
    }
    else if (!strcmp(argv[i], "--method") && i + 1 < argc) // Which method should we use (the 'standard' method, or the 'fast' method)?
    {
      i++;
      if (!strcmp(argv[i], "naive"))
        arg_method = NAIVE;
      else if (!strcmp(argv[i], "fast"))
        arg_method = FAST;
      else
      {
        printf("# ERROR: invalid method '%s'\n", argv[i]);
        exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "--nodes") && i + 1 < argc) // How many nodes should the model have?
    {
      arg_nodes = atoi(argv[++i]);
      assert(arg_nodes > 0);
    }
    else if (!strcmp(argv[i], "--vector_length") && i + 1 < argc) // How many states should the model have (for synthetic experiments)?
    {
      arg_vector_length = atoi(argv[++i]);
      assert(arg_vector_length > 0);
    }
    else
    {
      printf("# ERROR: invalid option '%s'\n", argv[i]);
      exit(EXIT_FAILURE);
    }
  }

  // Run experiment(s)
  run_experiment(arg_experiment, arg_method);

  return EXIT_SUCCESS;
}
