#include "matrixproduct.hpp"

/**
 Verify that both the slow and the fast solution produce the same outputs.

 N: matrix size.
 which: algorithm choice (see algos.hpp).
 The number of entries read is returned.
 */
int test(int N, int which)
{
  int evaluations = N * N * N;
  Scalar** A = new Scalar*[N];
  Scalar** B = new Scalar*[N];

  for (int i = 0; i < N; i++)
  {
    A[i] = new Scalar[N];
    B[i] = new Scalar[N];

    for (int j = 0; j < N; j++)
    {
      A[i][j] = rand() % N;
      B[i][j] = rand() % N;
    }
  }

  Scalar** C;
  if (which == NAIVE)
    C = slow(A, B, N);
  else
    C = sorting(A, B, which, N, evaluations);

  Scalar** Cslow = slow(A, B, N);
  for (int i = 0; i < N; i++)
  {
    for (int j = 0; j < N; j++)
    {
      if (C[i][j] != Cslow[i][j])
      {
        printf("%f %f\n", C[i][j], Cslow[i][j]);
        exit(1);
      }
      assert(C[i][j] == Cslow[i][j]);
    }
  }
  for (int i = 0; i < N; i++)
  {
    delete[] A[i];
    delete[] B[i];
    delete[] Cslow[i];
    delete[] C[i];
  }

  delete[] A;
  delete[] B;
  delete[] Cslow;
  delete[] C;

  return evaluations;
}

/**
 The naive solution.

 A, B: input matrices.
 N: matrix size.
 */
Scalar** slow(Scalar** A, Scalar** B, int N)
{
  Scalar** C = new Scalar *[N];
  for (int i = 0; i < N; i++)
    C[i] = new Scalar[N];

  for (int i = 0; i < N; i++)
    for (int k = 0; k < N; k++)
    {
      Scalar max = A[i][0] PLUS B[0][k];
      int maxind = 0;

      for (int j = 1; j < N; j++)
      {
        Scalar t = A[i][j] PLUS B[j][k];
        if (t COMP max)
        {
          max = t;
          maxind = j;
        }
      }
      C[i][k] = max;
    }

  return C;
}

/**
 The solution from our arXiv paper.

 A, B: input matrices.
 N: matrix size.
 evaluations: returns the number of addition operations.
 */
Scalar** sorting(Scalar** A, Scalar** B, int method, int N, int& evaluations)
{
  evaluations = 0;

  // Compute the permutations that sort the two matrices.
  Scalar** Btrans = new Scalar*[N];

  int** Pa = new int*[N];
  int** Pb = new int*[N];

  int** Painv = new int*[N];
  int** Pbinv = new int*[N];

  Scalar** C = new Scalar*[N];

  for (int i = 0; i < N; i++)
  {
    Btrans[i] = new Scalar[N];
    Pa[i] = new int[N];
    Pb[i] = new int[N];
    Painv[i] = new int[N];
    Pbinv[i] = new int[N];
    C[i] = new Scalar[N];
  }

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
      Btrans[i][j] = B[j][i];

  for (int i = 0; i < N; i++)
  {
    permfunction(A[i], Pa[i], Painv[i], N);
    permfunction(Btrans[i], Pb[i], Pbinv[i], N);
  }

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
    {
      int e = 0;
      // The inner loop: call the appropriate maximisation algorithm.
      if (method == NAIVE)
        C[i][j] = naive(A[i], Btrans[j], N, e);
      else if (method == FAST)
        C[i][j] = maximize(A[i], Btrans[j], Pa[i], Pb[j], Painv[i], Pbinv[j], e);
      else
      {
        printf("Unrecognized method: %d\n", method);
        exit(1);
      }
      evaluations += e;
    }

  // Cleanup.
  for (int i = 0; i < N; i++)
  {
    delete[] Btrans[i];
    delete[] Pa[i];
    delete[] Pb[i];
    delete[] Painv[i];
    delete[] Pbinv[i];
  }
  delete[] Btrans;
  delete[] Pa;
  delete[] Pb;
  delete[] Painv;
  delete[] Pbinv;

  return C;
}
