#include "textprocess.hpp"
#include "bprop.hpp"
#include "grid.hpp"

/** Unary potential for text-denoising. */
Scalar* unary_text(wchar_t obs, map<wchar_t, int> alphabet, Scalar prob)
{
  int N = alphabet.size();
  Scalar* res = new Scalar[N];
  for (int i = 0; i < N; i++)
  { // Just an indicator function testing if the output is correct.
    if (alphabet.find(obs) != alphabet.end() and alphabet[obs] == i)
      res[i] = -log(1 - prob);
    else
      res[i] = -log(prob);
  }
  return res;
}

/** Pairwise potentials for text-denoising. */
Scalar** text_prior(wchar_t* corpus, map<wchar_t, int> alphabet, int P, int skip, int skipchain)
{
  int N = alphabet.size();
  Scalar** prior = new Scalar*[N];
  for (int i = 0; i < N; i++)
  {
    prior[i] = new Scalar[N];
    for (int j = 0; j < N; j++)
      prior[i][j] = 0;
  }

  // Just the pairwise statistics of the text.
  for (int i = 0; i < P - skip; i++)
    prior[alphabet[corpus[i]]][alphabet[corpus[i + skip]]]++;

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
    {
      prior[i][j] = prior[i][j] / (P - skip);

      if (prior[i][j] == 0)
        prior[i][j] = 6;
      else
      {
        prior[i][j] = 4.0 / (1.0 + 4 * prior[i][j]); // Sigmoid function penalizes unreliable outputs.
      }
      if (skip == 2 and !skipchain) prior[i][j] = 0;
    }

  return prior;
}

int alphabet_size(const char* corpus, int P, int L, const char* enc)
{
  FILE* f = fopen(corpus, enc);
  if (!f)
  {
    printf("# Couldn't open %s\n", corpus);
    exit(1);
  }
  map<wchar_t, int> alphabet;
  wchar_t c;
  int ind = 0;
  int charsread = 0;
  while ((c = fgetwc(f)) != EOF and charsread < P + L)
  {
    // Skip past tabulators and line numbers in the text files.
    if (c == '\n')
    {
      while (c != '\t')
        c = fgetwc(f);
      c = ' ';
    }
    charsread++;
    if (alphabet.find(c) == alphabet.end())
    {
      alphabet[c] = ind;
      ind++;
    }
  }
  assert(charsread == P + L);
  fclose(f);

  return (int) alphabet.size();
}

/** Text-denoising experiment. */
int* textprocess(int method, const char* corpus, int P, int L, const char* enc, double& time, long long& nread, Scalar noise, int skipchain, int& errs, int& perf, Scalar& energy)
{
  int X = L;
  int* res = new int[X];

  srand(0);
  nread = 0;
  FILE* f = fopen(corpus, enc);
  map<wchar_t, int> alphabet;
  map<int, wchar_t> ralphabet;
  vector<wchar_t> valphabet;
  wchar_t c;
  int ind = 0;
  int charsread = 0;
  wchar_t* train = new wchar_t[P + 1];
  wchar_t* test = new wchar_t[L + 1];
  train[P] = '\0';
  train[L] = '\0';
  while ((c = fgetwc(f)) != EOF and charsread < P + L)
  {
    // Skip past tabulators and line numbers in the text files.
    if (c == '\n')
    {
      while (c != '\t')
        c = fgetwc(f);
      c = ' ';
    }
    if (charsread < P)
      train[charsread] = c;
    else
      test[charsread - P] = c;
    charsread++;
    if (alphabet.find(c) == alphabet.end())
    {
      alphabet[c] = ind;
      ralphabet[ind] = c;
      ind++;
      valphabet.push_back(c);
    }
  }
  assert(charsread == P + L);
  fclose(f);

  X = L;
  int Y = 1;

  Scalar** smoothness1 = text_prior(train, alphabet, P, 1, skipchain);
  Scalar** smoothness2 = text_prior(train, alphabet, P, 2, skipchain);

  wchar_t* original = new wchar_t[L + 1];
  memcpy(original, test, (L + 1) * sizeof(wchar_t));


  // Add noise to the text.
  for (int i = 0; i < L; i++)
    if ((rand() / (1.0 * RAND_MAX)) < noise) test[i] = valphabet[rand() % valphabet.size()];

  Scalar*** p = new Scalar**[Y];
  p[0] = new Scalar*[X];
  Scalar** node_prior = p[0];
  for (int i = 0; i < L; i++)
    node_prior[i] = unary_text(test[i], alphabet, noise);

  clock_t T = clock();

  if (!skipchain)
  {
    Scalar online, total, energy;
    int** res_g = grid(method, p, smoothness1, X, Y, alphabet.size(), online, total, energy, nread);
    for (int i = 0; i < X; i++)
      res[i] = res_g[0][i];
    delete[] res_g[0];
    delete[] res_g;
  }
  else
  {
	vector<cliquemessage*> cliques;
	// Costruct a chain-structured graphical model.
	for (int m = 0; m < X - 2; m++)
	{
	  Scalar** A = new Scalar*[alphabet.size()];
	  Scalar** B = new Scalar*[alphabet.size()];
	  Scalar** C = new Scalar*[alphabet.size()];

	  for (int i = 0; i < (int) alphabet.size(); i++)
	  {
		A[i] = new Scalar[alphabet.size()];
		B[i] = new Scalar[alphabet.size()];
		C[i] = new Scalar[alphabet.size()];

		for (int j = 0; j < (int) alphabet.size(); j++)
		{
		  A[i][j] = node_prior[m][i] + smoothness1[i][j];
		  B[i][j] = smoothness2[i][j];
		  C[i][j] = 0;
		  if (m == X - 3) C[i][j] = node_prior[m + 1][i] + node_prior[m + 2][j] + smoothness2[i][j];
		}
	  }
	  cliques.push_back(new cliquemessage(A, B, C, alphabet.size()));
	}
    gmodel G(cliques, method);

    T = clock();

    G.passmessage_chain(nread);
    for (int i = 0; i < X - 3; i++)
      G.MAP(LEFT, alphabet.size(), res + i, i, nread);
    G.MAP(RIGHT, alphabet.size(), res + X - 2, X - 3, nread);

    energy = SEMIRING_ZERO;
	for (int i = 0; i < X - 2; i++)
	{
	  cliquemessage* cm = cliques[i];
	  energy = energy PLUS cm->A[res[i]][res[i + 1]] PLUS cm->B[res[i]][res[i + 2]] PLUS cm->C[res[i + 1]][res[i + 2]];
	}

    for (int i = 0; i < (int) cliques.size(); i++)
  	  delete cliques[i];
  }

  time = ((double) (clock() - T)) / CLOCKS_PER_SEC;

  int error1 = 0;
  int error2 = 0;

  for (int i = 0; i < L; i++)
    if (test[i] != original[i]) error1++;

  for (int i = 0; i < L; i++)
    if (ralphabet[res[i]] != original[i]) error2++;

  errs = error1;
  //printf("initial errors = %d, final errors = %d\n", error1, error2);

  // Cleanup.
  for (int i = 0; i < (int) alphabet.size(); i++)
  {
    delete[] smoothness1[i];
    delete[] smoothness2[i];
  }
  delete[] smoothness1;
  delete[] smoothness2;

  for (int y = 0; y < Y; y++)
  {
    for (int x = 0; x < X; x++)
      delete[] p[y][x];
    delete[] p[y];
  }
  delete[] p;

  delete[] train;
  delete[] test;
  delete[] original;

  perf = error2;
  return res;
}
